@@ -23,12 +23,15 @@
 # along with mpop.  If not, see <http://www.gnu.org/licenses/>.
 """Module for testing the image and xrimage modules."""
 import os
-import sys
 import random
-import unittest
+import sys
 import tempfile
+import unittest
+from collections import OrderedDict
 from tempfile import NamedTemporaryFile
+
 import numpy as np
+
 from trollimage import image
 
 try:
@@ -958,7 +961,7 @@ class TestXRImage(unittest.TestCase):
             delay = img.save(tmp.name, compute=False)
             self.assertIsInstance(delay, tuple)
             self.assertIsInstance(delay[0], da.Array)
-            self.assertIsInstance(delay[1], xrimage.RIOFile)
+            self.assertIsInstance(delay[1], xrimage.RIODataset)
             da.store(*delay)
             delay[1].close()
 
@@ -1031,7 +1034,7 @@ class TestXRImage(unittest.TestCase):
             delay = img.save(tmp.name, compute=False)
             self.assertIsInstance(delay, tuple)
             self.assertIsInstance(delay[0], da.Array)
-            self.assertIsInstance(delay[1], xrimage.RIOFile)
+            self.assertIsInstance(delay[1], xrimage.RIODataset)
             da.store(*delay)
             delay[1].close()
 
@@ -1198,6 +1201,44 @@ class TestXRImage(unittest.TestCase):
             with rio.open(tmp.name) as f:
                 self.assertEqual(len(f.overviews(1)), 2)
 
+    @unittest.skipIf(sys.platform.startswith('win'), "'NamedTemporaryFile' not supported on Windows")
+    def test_save_tags(self):
+        """Test saving geotiffs with tags."""
+        import xarray as xr
+        from trollimage import xrimage
+        import rasterio as rio
+
+        # numpy array image
+        data = xr.DataArray(np.arange(75).reshape(5, 5, 3), dims=[
+            'y', 'x', 'bands'], coords={'bands': ['R', 'G', 'B']})
+        img = xrimage.XRImage(data)
+        tags = {'avg': img.data.mean(), 'current_song': 'disco inferno'}
+        self.assertTrue(np.issubdtype(img.data.dtype, np.integer))
+        with NamedTemporaryFile(suffix='.tif') as tmp:
+            img.save(tmp.name, tags=tags)
+            tags['avg'] = '37.0'
+            with rio.open(tmp.name) as f:
+                self.assertEqual(f.tags(), tags)
+
+    @unittest.skipIf(sys.platform.startswith('win'), "'NamedTemporaryFile' not supported on Windows")
+    def test_save_scale_offset(self):
+        """Test saving geotiffs with tags."""
+        import xarray as xr
+        from trollimage import xrimage
+        import rasterio as rio
+
+        data = xr.DataArray(np.arange(25).reshape(5, 5, 1), dims=[
+            'y', 'x', 'bands'], coords={'bands': ['L']})
+        img = xrimage.XRImage(data)
+        img.stretch()
+        with NamedTemporaryFile(suffix='.tif') as tmp:
+            img.save(tmp.name, include_scale_offset_tags=True)
+            tags = {'scale': 24.0 / 255, 'offset': 0}
+            with rio.open(tmp.name) as f:
+                ftags = f.tags()
+                for key, val in tags.items():
+                    self.assertAlmostEqual(float(ftags[key]), val)
+
     def test_gamma(self):
         """Test gamma correction."""
         import xarray as xr
@@ -1612,6 +1653,18 @@ class TestXRImage(unittest.TestCase):
             self.assertTrue(img2.mode == 'RGBA')
             self.assertTrue(len(img2.data.coords['bands']) == 4)
 
+    def test_final_mode(self):
+        """Test final_mode."""
+        import xarray as xr
+        from trollimage import xrimage
+
+        # numpy array image
+        data = xr.DataArray(np.arange(75).reshape(5, 5, 3), dims=[
+            'y', 'x', 'bands'], coords={'bands': ['R', 'G', 'B']})
+        img = xrimage.XRImage(data)
+        self.assertEqual(img.final_mode(None), 'RGBA')
+        self.assertEqual(img.final_mode(0), 'RGB')
+
     def test_colorize(self):
         """Test colorize with an RGB colormap."""
         import xarray as xr
@@ -1875,6 +1928,42 @@ class TestXRImage(unittest.TestCase):
             img.show()
             s.assert_called_once()
 
+    def test_apply_pil(self):
+        """Test the apply_pil method."""
+        import xarray as xr
+        from trollimage import xrimage
+
+        np_data = np.arange(75).reshape(5, 5, 3) / 75.
+        data = xr.DataArray(np_data, dims=[
+            'y', 'x', 'bands'], coords={'bands': ['R', 'G', 'B']})
+
+        dummy_args = [(OrderedDict(), ), {}]
+
+        def dummy_fun(pil_obj, *args, **kwargs):
+            dummy_args[0] = args
+            dummy_args[1] = kwargs
+            return pil_obj
+
+        img = xrimage.XRImage(data)
+        pi = mock.MagicMock()
+        img.pil_image = pi
+        res = img.apply_pil(dummy_fun, 'RGB')
+        # check that the pil image generation is delayed
+        pi.assert_not_called()
+        # make it happen
+        res.data.data.compute()
+        pi.return_value.convert.assert_called_with('RGB')
+
+        img = xrimage.XRImage(data)
+        pi = mock.MagicMock()
+        img.pil_image = pi
+        res = img.apply_pil(dummy_fun, 'RGB',
+                            fun_args=('Hey', 'Jude'),
+                            fun_kwargs={'chorus': "La lala lalalala"})
+        self.assertEqual(dummy_args, [({}, ), {}])
+        res.data.data.compute()
+        self.assertEqual(dummy_args, [(OrderedDict(), 'Hey', 'Jude'), {'chorus': "La lala lalalala"}])
+
 
 def suite():
     """Create the suite for test_image."""
